from __future__ import annotations

import functools
import re
from collections import OrderedDict as odict
from collections.abc import Iterator
from typing import Any, TypedDict, cast

from annet.annlib import lib
from annet.vendors import tabparser


# =====
def _merge_trees(t1: odict[Any, Any], t2: odict[Any, Any]) -> odict[Any, Any]:
    if not t1:
        return t2
    if not t2:
        return t1
    ret = t1.copy()
    for k, v in t2.items():
        if k in ret:
            ret[k]["children"] = _merge_trees(ret[k]["children"], v["children"])
        else:
            ret[k] = v
    return ret


def _convert(tree: ParsedTree) -> odict[Any, Any]:
    ret: odict[Any, Any] = odict()
    for rule_id, attrs in tree:
        if rule_id not in ret:
            ret[rule_id] = attrs | {"children": odict()}
        ret[rule_id]["children"] = _merge_trees(ret[rule_id]["children"], _convert(attrs["children"]))
    return ret


def parse_text(text: str, params_scheme) -> odict[Any, Any]:
    ret = _parse_tree_with_params(tabparser.parse_to_tree_multi(text, _split_rows, ["#"]), params_scheme)
    return _convert(ret)


class _ParsedTreeNode(TypedDict):
    row: str
    type: str
    params: dict[str, Any]
    children: ParsedTree
    raw_rule: str
    context: Any


ParsedTree = list[tuple[str, _ParsedTreeNode]]


def parse_text_multi(text: str, params_scheme) -> ParsedTree:
    ret = _parse_tree_with_params(tabparser.parse_to_tree_multi(text, _split_rows, ["#"]), params_scheme)
    return ret


@functools.lru_cache()
def compile_row_regexp(row, flags=0):
    if "(?i)" in row:
        row = row.replace("(?i)", "")
        flags |= re.IGNORECASE

    if "*" in row:
        row = re.sub(r"\(([^\?])", r"(?:\1", row)  # Все дефолтные группы превратить в non-captured
        row = re.sub(r"\*/(\S+)/", r"(\1)", row)  # */{regex_one_word}/ -> ({regex_one_word})
        row = re.sub(r"(^|\s)\*", r"\1([^\\s]+)", row)

    # Заменяем <someting> на named-группы
    row = re.sub(r"<(\w+)>", r"(?P<\1>\\w+)", row)

    if row.endswith("~"):
        # We determine the most specific regex for the row at matching in match_row_to_acls
        row = row[:-1] + "(.+)"
    elif row.endswith("..."):
        row = row[:-3]
    elif "~/" in row:
        # ~/{regex}/ -> {regex}, () не нужны поскольку уже (?:) - non-captured
        row = re.sub(r"~/(((?!~/).)+)/", r"\1", row)
    else:
        row += r"(?:\s|$)"
    row = re.sub(r"\s+", r"\\s+", row)
    return re.compile("^" + row, flags=flags)


# =====
def _split_rows(text: str) -> Iterator[str]:
    for row in re.split(r"\n(?!\s*%(?!context))", text):
        yield row.replace("\n", " ")


def _parse_tree_with_params(raw_tree: tabparser.SimpleTree, scheme, context: dict | None = None) -> ParsedTree:
    tree: ParsedTree = []
    if context is None:
        context = {}
    for raw_rule, children in raw_tree:
        (row, params) = _parse_raw_rule(raw_rule, scheme)
        row_type = "normal"

        if row.startswith("!"):
            row = row[1:].strip()
            if len(row) == 0:
                continue
            row_type = "ignore"
        elif context_raw := params.get("context"):
            context = _parse_context(context, context_raw)
            continue
        tree.append(
            (
                raw_rule,
                {
                    "row": row,
                    "type": row_type,
                    "params": params,
                    "children": _parse_tree_with_params(children, scheme, cast(dict, context).copy()),
                    "raw_rule": raw_rule,
                    "context": cast(dict, context).copy(),
                },
            )
        )
    return tree


def _parse_raw_rule(raw_rule: str, scheme) -> tuple[str, dict[str, str]]:
    params: dict[str, str] = {}

    row, *params_raw = re.split(r"(?:^|\s)%(?=[a-zA-Z_]\w*)", raw_rule)
    for param in params_raw:
        name, _, value = param.partition("=")
        params[name.strip()] = value.strip() or "1"

    row = re.sub(r"\s+", " ", row.strip())
    params = _fill_and_validate(params, scheme, raw_rule)
    return row, params


def _fill_and_validate(params, scheme, raw_rule):
    return {
        key: (
            attrs["validator"](params[key])
            if key in params
            else (attrs["default"](raw_rule) if callable(attrs["default"]) else attrs["default"])
        )
        for (key, attrs) in scheme.items()
    }


def match_context(ifcontext, context):
    if not ifcontext:
        return True
    for ifcontext_value in ifcontext:
        name, value = ifcontext_value.split(":")
        if name in context:
            if context[name] == value:
                return True
    return False


def _parse_context(context, row):
    name, value = row.strip().split(":")
    return lib.merge_dicts(context, {name: value})



import itertools
from typing import List

from annet.generators import BaseGenerator
from annet.storage import Storage

from . import hostname, lldp


def get_generators(store: Storage) -> List[BaseGenerator]:
    return list(
        itertools.chain.from_iterable(
            [
                hostname.get_generators(store),
                lldp.get_generators(store),
            ]
        )
    )



from annet.annlib.types import Op
from annet.rulebook import common


def diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    for iface_row in old:
        _filter_channel_members(old[iface_row])
    for iface_row in new:
        _filter_channel_members(new[iface_row])

    ret = common.default_diff(old, new, diff_pre, _pops)
    vpn_changed = False
    for op, cmd, _, _ in ret:
        if op in {Op.ADDED, Op.REMOVED}:
            vpn_changed |= is_vpn_cmd(cmd)
    if vpn_changed:
        for cmd in list(old.keys()):
            if is_ip_cmd(cmd) and not is_vpn_cmd(cmd):
                del old[cmd]
        ret = common.default_diff(old, new, diff_pre, _pops)
    return ret


# ===

# Вырезает все команды не разрешенные
# на членах агрегата. В running-config
# листинге они наследуются от самого port-channel


def _filter_channel_members(tree):
    if any(is_in_channel(x) for x in tree):
        for cmd in list(tree.keys()):
            if not _is_allowed_on_channel(cmd):
                del tree[cmd]


def is_in_channel(cmd_line):
    """
    Признак того, что это lagg member
    """
    return cmd_line.startswith("channel-group")


# Возможно тут есть еще какие-то команды
def _is_allowed_on_channel(cmd_line):
    return cmd_line.startswith(
        (
            "channel-group",
            "cdp",
            "description",
            "inherit",
            "ip port",
            "ipv6 port",
            "mac port",
            "lacp",
            "switchport host",
            "shutdown",
            "rate-limit cpu",
            "snmp trap link-status",
        )
    )


def is_vpn_cmd(cmd):
    return cmd.startswith("vrf member")


def is_ip_cmd(cmd):
    return cmd.startswith(("ip ", "ipv6 "))


def mtu(rule, key, diff, **kwargs):
    """
    Удаляем mtu без указания значения
    """
    if diff[Op.REMOVED]:
        yield (False, "no mtu", None)
    elif diff[Op.ADDED]:
        yield from common.default(rule, key, diff, **kwargs)


def description(rule, key, diff, **kwargs):
    """
    Удаляем description без указания значения
    """
    if diff[Op.REMOVED]:
        yield (False, "no description", None)
    elif diff[Op.ADDED]:
        yield from common.default(rule, key, diff, **kwargs)


def sflow(rule, key, diff, **kwargs):
    """
    Команда sflow sampling-rate * direction ingress max-header-size *
    сносится без указания sampling-rate и max-header-size
    """
    if diff[Op.REMOVED]:
        if "ingress" in diff[Op.REMOVED][0]["row"]:
            yield (False, "no sflow sampling-rate direction ingress", None)
        elif "egress" in diff[Op.REMOVED][0]["row"]:
            yield (False, "no sflow sampling-rate direction egress", None)
    else:
        yield from common.default(rule, key, diff, **kwargs)


def lldp(rule, key, diff, **kwargs):
    """
    Обрабатываем блок lldp-agent
    """
    result = common.default(rule, key, diff, **kwargs)
    for op, cmd, ch in result:
        # Не удаляем все что начинается с set, т.к. set перезаписывает предыдущий конфиг
        if diff[Op.REMOVED] and "set lldp" in cmd:
            pass
        # В случае lldp tlv ... select удаляем все что до select
        elif diff[Op.REMOVED] and cmd.endswith("select"):
            yield (op, " ".join(cmd.split()[:-1]), ch)
        else:
            yield (op, cmd, ch)



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional

import annetbox.v42.models

from annet.adapters.netbox.common.models import Entity, Interface, InterfaceType, IpAddress, IpFamily, Label, Prefix
from annet.adapters.netbox.v41.models import FHRPGroupAssignmentV41, InterfaceV41, IpAddressV41, NetboxDeviceV41


@dataclass
class PrefixV42(Prefix):
    scope: Optional[Entity] = None
    scope_type: str | None = None

    @property
    def site(self) -> Optional[Entity]:
        if self.scope_type == "dcim.site":
            return self.scope
        return None


@dataclass
class IpAddressV42(IpAddress[PrefixV42]):
    role: Optional[Label] = None
    prefix: Optional[PrefixV42] = None


@dataclass
class InterfaceV42(Interface[IpAddressV42, FHRPGroupAssignmentV41]):
    def _add_new_addr(self, address_mask: str, vrf: Entity | None, family: IpFamily) -> None:
        self.ip_addresses.append(
            IpAddressV42(
                id=0,
                display=address_mask,
                address=address_mask,
                vrf=vrf,
                prefix=None,
                family=family,
                created=datetime.now(timezone.utc),
                last_updated=datetime.now(timezone.utc),
                tags=[],
                status=Label(value="active", label="Active"),
                assigned_object_id=self.id,
            )
        )


@dataclass
class NetboxDeviceV42(NetboxDeviceV41):
    def __hash__(self):
        return hash((self.id, type(self)))

    def _make_interface(self, name: str, type: InterfaceType) -> InterfaceV42:
        return InterfaceV42(
            name=name,
            device=self,
            enabled=True,
            description="",
            type=type,
            id=0,
            vrf=None,
            display=name,
            untagged_vlan=None,
            tagged_vlans=[],
            ip_addresses=[],
            connected_endpoints=[],
            mode=None,
        )


# should be unified in case of netbox update
Vrf = annetbox.v42.models.Vrf
Vlan = annetbox.v42.models.Vlan



from annet.annlib.types import Op
from annet.rulebook import common


def undo_peer_group(rule, key, diff, **_):
    """Correctly removes neighbors that were added to the peer-group."""
    for action in diff[Op.ADDED]:
        yield (True, action["row"], None)

    if diff[Op.REMOVED]:
        is_group = any(action["row"].endswith(" peer-group") for action in diff[Op.REMOVED])

        group_name = ""
        for action in diff[Op.REMOVED]:
            if " peer-group " in action["row"]:
                group_name = action["row"].split()[-1]
        is_remote_as = any(" remote-as " in action["row"] for action in diff[Op.REMOVED])

        is_group_in_removed = False
        if _["rule_pre"]["items"].get(tuple([group_name])):
            for action in _["rule_pre"]["items"][tuple([group_name])]["removed"]:
                if action["row"].endswith(f"{group_name} peer-group"):
                    is_group_in_removed = True
                    break

        for action in diff[Op.REMOVED]:
            row = action["row"]
            if is_group:
                if row.endswith(" peer-group"):
                    yield (False, "no " + row, None)
            elif is_group_in_removed:
                continue
            elif is_remote_as or group_name:
                if " peer-group " in row:
                    yield (False, "no " + row, None)
            else:
                yield (False, "no " + row, None)


def undo_peer(rule, key, diff, **kwargs):
    """
    If we remove a neighbor, we just remove configuration with remote-as command.
    But if we remove specific neighbor's options, without neighbor deletion
    we need check if neighbor * remote-as not exists in Op.REMOVED rule_pre items.
    """
    if diff[Op.REMOVED]:
        if "remote-as" in diff[Op.REMOVED][0]["row"]:
            yield (False, "no " + diff[Op.REMOVED][0]["row"], None)
        else:
            is_neighbor_removing = False
            for item in kwargs["rule_pre"]["items"].values():
                if item[Op.REMOVED]:
                    if f"neighbor {key[0].split()[0]} remote-as" in item[Op.REMOVED][0]["row"]:
                        is_neighbor_removing = True
                        break
            if not is_neighbor_removing:
                yield from common.default(rule, key, diff, **kwargs)
    else:
        yield from common.default(rule, key, diff, **kwargs)



from __future__ import annotations

import functools
import re
from collections.abc import Iterator
from typing import Any, TypedDict

from valkit.common import valid_bool, valid_string_list

from annet.vendors import registry_connector

from . import syntax


# =====
# 'global' is a keyword, so we cant use normal TypedDict declaration
_CompiledOrderingAttrs = TypedDict(
    "_CompiledOrderingAttrs",
    {
        "direct_regexp": re.Pattern[str],
        "reverse_regexp": re.Pattern[str],
        "order_reverse": bool,
        "global": bool,  # TODO: rename to something else so that it is not a keyword
        "scope": list[str] | None,
        "raw_rule": str,
        "context": Any,
        "split": bool,
    },
)


class _CompiledOrderingItem(TypedDict):
    attrs: _CompiledOrderingAttrs
    children: CompiledTree


CompiledTree = list[tuple[str, _CompiledOrderingItem]]


@functools.lru_cache()
def compile_ordering_text(text: str, vendor: str):
    return _compile_ordering(
        tree=syntax.parse_text_multi(
            text,
            params_scheme={
                "order_reverse": {
                    "validator": valid_bool,
                    "default": False,
                },
                "global": {
                    "validator": valid_bool,
                    "default": False,
                },
                "scope": {
                    "validator": valid_string_list,
                    "default": None,
                },
                "split": {
                    "validator": valid_bool,
                    "default": False,
                },
            },
        ),
        reverse_prefix=registry_connector.get()[vendor].reverse,
    )


def decompile_ordering_rulebook(rb: CompiledTree) -> str:
    def _decompile_ordering_text(rb: CompiledTree, level: int) -> Iterator[str]:
        indent = "  "
        for _, attrs in rb:
            yield indent * level + attrs["attrs"]["raw_rule"]
            yield from _decompile_ordering_text(attrs["children"], level + 1)

    return "\n".join(_decompile_ordering_text(rb, 0))


# =====


def _compile_ordering(tree: syntax.ParsedTree, reverse_prefix: str) -> CompiledTree:
    ordering: CompiledTree = []
    for rule_id, attrs in tree:
        if attrs["type"] == "normal":
            ordering.append(
                (
                    rule_id,
                    {
                        "attrs": _CompiledOrderingAttrs(
                            {
                                "direct_regexp": syntax.compile_row_regexp(attrs["row"]),
                                "reverse_regexp": (
                                    syntax.compile_row_regexp(reverse_prefix + " " + attrs["row"])
                                    if not attrs["row"].startswith(reverse_prefix + " ")
                                    else syntax.compile_row_regexp(
                                        re.sub(r"^%s\s+" % (reverse_prefix), "", attrs["row"])
                                    )
                                ),
                                "order_reverse": attrs["params"]["order_reverse"],
                                "global": attrs["params"]["global"],
                                "scope": attrs["params"]["scope"],
                                "raw_rule": attrs["raw_rule"],
                                "context": attrs["context"],
                                "split": attrs["params"]["split"],
                            }
                        ),
                        "children": _compile_ordering(attrs["children"], reverse_prefix),
                    },
                )
            )
    return ordering



import dataclasses
from dataclasses import dataclass, fields
from typing import Any, Iterable, List, Optional, Sequence

import yaml

from annet.adapters.netbox.common.manufacturer import get_breed
from annet.annlib.netdev.views.dump import DumpableView
from annet.annlib.netdev.views.hardware import HardwareView
from annet.connectors import AdapterWithName
from annet.hardware import hardware_connector
from annet.storage import Device as DeviceProtocol
from annet.storage import Query, Storage, StorageProvider


@dataclass
class Interface(DumpableView):
    name: str
    description: str
    enabled: bool = True

    @property
    def _dump__list_key(self):
        return self.name


@dataclass
class DeviceStorage:
    fqdn: str

    vendor: Optional[str] = None
    hw_model: Optional[str] = None
    sw_version: Optional[str] = None
    breed: Optional[str] = None

    hostname: Optional[str] = None
    serial: Optional[str] = None
    id: Optional[str] = None
    interfaces: Optional[list[Interface]] = None
    storage: Optional[Storage] = None

    hw: HardwareView = dataclasses.field(init=False)

    def __post_init__(self):
        if not self.id:
            self.id = self.fqdn
        if not self.hostname:
            self.hostname = self.fqdn.split(".")[0]

        if self.hw_model:
            hw = HardwareView(self.hw_model, self.sw_version)
            if self.vendor and self.vendor != hw.vendor:
                raise Exception(f"Vendor {self.vendor} is not vendor from hw model ({hw.vendor})")
            else:
                self.vendor = hw.vendor
        else:
            hw_provider = hardware_connector.get()
            hw: HardwareView = hw_provider.vendor_to_hw(self.vendor)
            if not hw:
                raise Exception("unknown vendor")
            self.hw_model = hw.model
        self.hw = hw

        if not self.breed:
            if self.hw.model:
                parts = self.hw.model.split(maxsplit=1)
                if len(parts) >= 2:
                    self.breed = get_breed(parts[0], parts[1])
            if not self.breed:
                self.breed = self.hw.vendor

        if isinstance(self.interfaces, list):
            interfaces = []
            for iface in self.interfaces:
                try:
                    interfaces.append(Interface(**iface))
                except Exception as e:
                    raise Exception("unable to parse %s as Interface: %s" % (iface, e))
            self.interfaces = interfaces

    def set_storage(self, storage: Storage):
        self.storage = storage


@dataclass
class Device(DeviceProtocol, DumpableView):
    dev: DeviceStorage

    def __hash__(self):
        return hash((self.id, type(self)))

    def __eq__(self, other):
        return type(self) is type(other) and self.fqdn == other.fqdn and self.vendor == other.vendor

    def is_pc(self) -> bool:
        return False

    @property
    def hostname(self) -> str:
        return self.dev.hostname

    @property
    def fqdn(self) -> str:
        return self.dev.fqdn

    @property
    def id(self):
        return self.dev.id

    @property
    def storage(self) -> Storage:
        return self.dev.storage

    @property
    def hw(self) -> HardwareView:
        return self.dev.hw

    @property
    def breed(self) -> str:
        return self.dev.breed

    @property
    def neighbours_ids(self):
        return []

    @property
    def neighbours_fqdns(self) -> list[str]:
        return []

    def make_lag(self, lag: int, ports: Sequence[str], lag_min_links: Optional[int]) -> Interface:
        raise NotImplementedError

    def add_svi(self, svi: int) -> Interface:
        raise NotImplementedError

    def add_subif(self, interface: str, subif: int) -> Interface:
        raise NotImplementedError

    def find_interface(self, name: str) -> Optional[Interface]:
        raise NotImplementedError

    def flush_perf(self):
        pass


@dataclass
class Devices:
    devices: list[Device]

    def __post_init__(self):
        if isinstance(self.devices, list):
            devices = []
            for dev in self.devices:
                try:
                    devices.append(Device(dev=DeviceStorage(**dev)))
                except Exception as e:
                    raise Exception("unable to parse %s as Device: %s" % (dev, e))
            self.devices = devices


class Provider(StorageProvider, AdapterWithName):
    def storage(self):
        return storage_factory

    def opts(self):
        return StorageOpts

    def query(self):
        return Query

    @classmethod
    def name(cls) -> str:
        return "file"


@dataclass
class Query(Query):
    query: List[str]

    @classmethod
    def new(cls, query: str | Iterable[str], hosts_range: Optional[slice] = None) -> "Query":
        if hosts_range is not None:
            raise ValueError("host_range is not supported")
        return cls(query=list(query))

    @property
    def globs(self):
        return self.query

    def is_empty(self) -> bool:
        return len(self.query) == 0


class StorageOpts:
    def __init__(self, path: str):
        self.path = path

    @classmethod
    def parse_params(cls, conf_params: Optional[dict[str, str]], cli_opts: Any):
        path = conf_params.get("path")
        if not path:
            raise Exception("empty path")
        return cls(path=path)


def storage_factory(opts: StorageOpts) -> Storage:
    return FS(opts)


class FS(Storage):
    def __init__(self, opts: StorageOpts):
        self.opts = opts
        self.inventory: Devices = read_inventory(opts.path, self)

    def __enter__(self):
        return self

    def __exit__(self, _, __, ___):
        pass

    def resolve_object_ids_by_query(self, query: Query) -> list[str]:
        result = filter_query(self.inventory.devices, query)
        return [dev.fqdn for dev in result]

    def resolve_fdnds_by_query(self, query: Query) -> list[str]:
        result = filter_query(self.inventory.devices, query)
        return [dev.fqdn for dev in result]

    def make_devices(
        self,
        query: Query | list,
        preload_neighbors=False,
        use_mesh=None,
        preload_extra_fields=False,
        **kwargs,
    ) -> list[Device]:
        if isinstance(query, list):
            query = Query.new(query)
        result = filter_query(self.inventory.devices, query)
        return result

    def get_device(self, obj_id: str, preload_neighbors=False, use_mesh=None, **kwargs) -> Device:
        result = filter_query(self.inventory.devices, Query.new(obj_id))
        if not result:
            raise Exception("not found")
        return result[0]

    def flush_perf(self):
        pass

    def resolve_all_fdnds(self) -> list[str]:
        return [d.fqdn for d in self.inventory.devices]

    def search_connections(self, device: "Device", neighbor: "Device") -> list[tuple["Interface", "Interface"]]:
        return []


def filter_query(devices: list[Device], query: Query) -> list[Device]:
    result: list[Device] = []
    for dev in devices:
        if dev.fqdn in query.query:
            result.append(dev)
    return result


def read_inventory(path: str, storage: Storage) -> Devices:
    with open(path, "r") as f:
        file_data = yaml.load(f, Loader=yaml.SafeLoader)
    res = dataclass_from_dict(Devices, file_data)
    for dev in res.devices:
        dev.dev.set_storage(storage)
    return res


def dataclass_from_dict(klass: type, d: dict[str, Any]):
    try:
        fieldtypes = {f.name: f.type for f in fields(klass)}
    except TypeError:
        return d
    return klass(**{f: dataclass_from_dict(fieldtypes[f], d[f]) for f in d})



from __future__ import annotations

from typing import Iterable, Union

from annet.lib import add_annotation, flatten

from .annotate import AbstractAnnotateFormatter, annotate_formatter_connector
from .base import NONE_SEARCHER, TreeGenerator, _filter_str
from .exceptions import InvalidValueFromGenerator


class PartialGenerator(TreeGenerator):
    TYPE = "PARTIAL"

    def __init__(self, storage):
        super().__init__()
        self.storage = storage
        self._running_gen = None
        self._annotate: AbstractAnnotateFormatter = annotate_formatter_connector.get(self)
        self._annotations = []

    def supports_device(self, device) -> bool:
        if self.__class__.run is PartialGenerator.run:
            return bool(self._get_vendor_func(device.hw.vendor, "run"))
        else:
            return True

    def acl(self, device):
        if acl_func := self._get_vendor_func(device.hw.vendor, "acl"):
            return acl_func(device)
        return None

    def acl_safe(self, device):
        if acl_func := self._get_vendor_func(device.hw.vendor, "acl_safe"):
            return acl_func(device)
        return None

    def run(self, device) -> Iterable[Union[str, tuple]] | None:
        if run_func := self._get_vendor_func(device.hw.vendor, "run"):
            return run_func(device)
        return None

    def get_user_runner(self, device):
        if self.__class__.run is not PartialGenerator.run:
            return self.run
        return self._get_vendor_func(device.hw.vendor, "run")

    def _get_vendor_func(self, vendor: str, func_name: str):
        attr_name = f"{func_name}_{vendor}"
        return getattr(self, attr_name, None)

    # =====

    def __call__(self, device, annotate: bool = False):
        self._indents = []
        self._rows = []

        self._running_gen = self.run(device)
        for text in self._running_gen:
            if isinstance(text, tuple):
                text = " ".join(map(_filter_str, flatten(text)))
            else:
                text = _filter_str(text)
            self._append_text(text)

        for row, annotation in zip(self._rows, self._annotations):
            if NONE_SEARCHER.search(row):
                raise InvalidValueFromGenerator("Found 'None' in yield result: %s" % add_annotation(row, annotation))

        if annotate:
            generated_rows = (add_annotation(x, y) for (x, y) in zip(self._rows, self._annotations))
        else:
            generated_rows = self._rows

        return "\n".join((*generated_rows, ""))

    def _append_text(self, text):
        def annotation_cb(row):
            self._annotations.append(self._annotate.make_annotation(self._running_gen))
            return row

        self._append_text_cb(text, row_cb=annotation_cb)

    def get_running_line(self) -> tuple[str, int]:
        return self._annotate.get_running_line(self._running_gen)

    @classmethod
    def literal(cls, item):
        return '"{}"'.format(item)

    def __repr__(self):
        return "<%s>" % self.__class__.__name__



__all__ = [
    "GlobalOptions",
    "MeshSession",
    "DirectPeer",
    "IndirectPeer",
    "MeshExecutor",
    "MeshRulesRegistry",
    "Left",
    "Right",
    "Match",
    "VirtualLocal",
    "VirtualPeer",
    "PortProcessor",
    "separate_ports",
    "united_ports",
]

from .executor import MeshExecutor
from .match_args import Left, Match, Right
from .port_processor import PortProcessor, separate_ports, united_ports
from .registry import (
    DirectPeer,
    GlobalOptions,
    IndirectPeer,
    MeshRulesRegistry,
    MeshSession,
    VirtualLocal,
    VirtualPeer,
)



import warnings
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional

from annet.adapters.netbox.common.models import (
    DeviceIp,
    Entity,
    EntityWithSlug,
    FHRPGroup,
    FHRPGroupAssignment,
    Interface,
    InterfaceType,
    IpAddress,
    IpFamily,
    Label,
    NetboxDevice,
    Prefix,
)


@dataclass
class PrefixV41(Prefix):
    site: Optional[Entity] = None


@dataclass
class IpAddressV41(IpAddress[PrefixV41]):
    role: Optional[Label] = None
    prefix: Optional[PrefixV41] = None


@dataclass
class DeviceIpV41(DeviceIp):
    family: IpFamily


@dataclass
class FHRPGroupV41(FHRPGroup[DeviceIpV41]):
    pass


@dataclass
class FHRPGroupAssignmentV41(FHRPGroupAssignment[FHRPGroupV41]):
    pass


@dataclass
class InterfaceV41(Interface[IpAddressV41, FHRPGroupAssignmentV41]):
    def _add_new_addr(self, address_mask: str, vrf: Entity | None, family: IpFamily) -> None:
        self.ip_addresses.append(
            IpAddressV41(
                id=0,
                display=address_mask,
                address=address_mask,
                vrf=vrf,
                prefix=None,
                family=family,
                created=datetime.now(timezone.utc),
                last_updated=datetime.now(timezone.utc),
                tags=[],
                status=Label(value="active", label="Active"),
                assigned_object_id=self.id,
            )
        )


@dataclass
class NetboxDeviceV41(NetboxDevice[InterfaceV41, DeviceIpV41]):
    role: EntityWithSlug

    @property
    def device_role(self):
        warnings.warn("'device_role' is deprecated, use 'role' instead.", DeprecationWarning, stacklevel=2)
        return self.role

    def __hash__(self):
        return hash((self.id, type(self)))

    def _make_interface(self, name: str, type: InterfaceType) -> InterfaceV41:
        return InterfaceV41(
            name=name,
            device=self,
            enabled=True,
            description="",
            type=type,
            id=0,
            vrf=None,
            display=name,
            untagged_vlan=None,
            tagged_vlans=[],
            ip_addresses=[],
            connected_endpoints=[],
            mode=None,
        )



from typing import Any, Dict, List

from annet.connectors import AdapterWithConfig
from annet.deploy import Fetcher
from annet.storage import Device


class StubFetcher(Fetcher, AdapterWithConfig):
    @classmethod
    def with_config(cls, **kwargs: Dict[str, Any]) -> Fetcher:
        return cls(**kwargs)

    async def fetch_packages(
        self,
        devices: list[Device],
        processes: int = 1,
        max_slots: int = 0,
    ) -> tuple[dict[Device, str], dict[Device, Any]]:
        raise NotImplementedError()

    async def fetch(
        self,
        devices: list[Device],
        files_to_download: dict[str, list[str]] | None = None,
        processes: int = 1,
        max_slots: int = 0,
    ):
        raise NotImplementedError()
